#include "adf4355.h"

ADF4355::ADF4355(uint32_t tcxo, uint64_t VCOMIN, uint64_t VCOMAX, uint32_t pfdTarget, uint8_t type) {
    this->tcxo = tcxo;
    this->type = type;
    this->VCOMIN = VCOMIN;
    this->VCOMAX = VCOMAX;

    /*
     * PFD Calculation to be validated
     *
     * Fpfd = refin*(1+D)/(R*(1+T))
     * R = refin*(1+D)/(Fpfd*(1+T))
     *
     */
    if(tcxo >= pfdTarget * 2) {
        this->mul2 = 0;
        this->div2 = 1;
    } else if(tcxo >= pfdTarget && tcxo < pfdTarget * 2) {
        this->mul2 = 0;
        this->div2 = 0;
    } else if(tcxo <= pfdTarget/2) {
        this->mul2 = 1;
        this->div2 = 0;
    } else if(tcxo < pfdTarget && tcxo > pfdTarget/2) {
        this->mul2 = 0;
        this->div2 = 0;
    }

    uint32_t tmpR = (double)tcxo * ((1+(double)this->mul2)/((double)pfdTarget*(1+(double)this->div2)));
    if(tmpR > 1023) {
        this->R = 1023;
    } else if(tmpR < 1) {
        this->R = 1;
    } else {
        this->R = tmpR;
    }

    this->PFD = ((double)tcxo * (double)((1+(double)this->mul2)/((double)this->R*(1+(double)this->div2))));
}

uint8_t * ADF4355::getBytes(uint16_t pointN) {
    return this->freqParams[pointN].getBlob();
}

uint8_t * ADF4355::getUpdateBytes(uint16_t pointN) {
    return this->freqParams[pointN].getUpdateBlob();
}

/* Calculate frequency range */
void ADF4355::initFreqRange(uint64_t startFreq, uint64_t stopFreq, uint16_t nbPoints){
    uint64_t freq = 0;

    uint64_t step = (nbPoints > 1) ? ((stopFreq - startFreq) / (nbPoints - 1)) : 0;
    this->freqParams = new ADFParam[nbPoints];
    this->freqs = new uint64_t[nbPoints];

    for(uint16_t i = 0; i < nbPoints; i++) {
        if(i == nbPoints-1)
            freq = stopFreq;
        else
            freq = startFreq + i * step;

        freqs[i] = freq;

        this->freqParams[i].setChannel(step);
        this->freqParams[i].FreqParams(freq, this->VCOMIN, this->VCOMAX, this->PFD, this->div2, this->mul2, this->R);
        this->freqParams[i].setRegs(this->type);

        this->freqParams[i].setBlobs(this->type);
    }
}

/*
 * Parameters Class
 */

ADFParam::ADFParam() {

}

ADFParam::ADFParam(uint32_t freq) {
    this->freq = freq;
}

ADFParam::ADFParam(uint32_t freq, uint32_t *regs) {
    this->freq = freq;
    for(int i = 0; i < 13; i++) {
        this->regs[i] = regs[i];
    }
}

void ADFParam::setReg(uint8_t regN, uint32_t regVal) {
    this->regs[regN] = regVal;
}

uint32_t ADFParam::getReg(uint8_t regN) {
    return this->regs[regN];
}

uint8_t * ADFParam::getBlob() {
    return this->fblob;
}

uint8_t * ADFParam::getUpdateBlob() {
    return this->fupblob;
}

void ADFParam::setFreq(uint32_t freq) {
    this->freq = freq;
}

uint64_t ADFParam::getCalcFreq() {
    return (((double)this->INT + ((double)this->FRAC1+((double)this->FRAC2/(double)this->MOD2))/(double)16777216)*(double)this->PFD)/(double)(1 << this->divOut);
}

void ADFParam::FreqParams(uint64_t freq, uint64_t VCOMIN, uint64_t VCOMAX, uint32_t PFD, uint8_t div2, uint8_t mul2, uint16_t R) {
    this->div2 = div2;
    this->mul2 = mul2;
    this->R = R;
    this->PFD = PFD;

    // Search VCO Frequency
    // Divider 1:2:4:...:64
    uint64_t vcoFreq = 0;
    this->upperOut = 0;

    // Handle the upper range !
    if(freq > VCOMAX) {
        this->upperOut = 1;
        freq = (uint64_t)((double)freq / (double)2);
    }

    for(uint8_t i = 0; i < 7; i++) {
        if((freq * (1 << i)) >= VCOMIN && (freq * (1 << i)) <= VCOMAX) {
            vcoFreq = freq * (1 << i);
            this->divOut = i;
            break;
        }
    }

    double N = (double)vcoFreq / (double)PFD;

    this->INT = (uint64_t)N;

    double FRAC = ((double)vcoFreq / (double)PFD) - (double)INT;

    this->MOD1 = 16777216;

    this->FRAC1 = FRAC * MOD1;

    //uint64_t a = this->channel;
    uint64_t a = this->PFD;
    uint64_t b = 200000;//this->PFD;

    if(this->PFD > this->channel) {
        a = this->PFD;
        b = 200000;//this->channel;
    }

    /* Search of GCD */
    uint64_t r = 0;
    do {
        r = a % b;
        if(r == 0) break;
        a = b;
        b = r;
    }while(1);

    this->MOD2 = (this->PFD / b);// & (uint64_t)0x3FFF;
    this->FRAC2 = (uint64_t)(((N-(double)INT) * (double)this->MOD1 - (double)FRAC1) * MOD2) & (uint64_t)0x3FFF;

}

bool ADFParam::isFreqUpper() {
    return (this->upperOut == 1);
}

void ADFParam::setChannel(uint64_t step) {
    this->channel = step;
}

/* Incomplete implementation to be tested ! */
void ADFParam::setRegs(uint8_t type) {
    this->regs[0] = 1 << 21 | INT << 4;
    this->regs[1] = FRAC1 << 4 | 1;
    this->regs[2] = FRAC2 << 18 | MOD2 << 4 | 2;
    this->regs[3] = 3;
    this->regs[4] = 6 << 27 | this->mul2 << 26 | this->div2 << 25 | this->R << 15 | 2 << 10 | 1 << 9 | 1 << 8 | 1 << 7 | 4;
    this->regs[5] = 0x00800025;

    uint8_t ibleed = ((4*0.0009) / (double)INT)/(3.75*1e-6);

    if(type == 0) { // ADF4355
        this->regs[6] = 1 << 29 | 0xA << 25 | 1 << 24 | (this->divOut) << 21 | ibleed << 13 | 1 << 9 | 1 << 6 | 3 << 4 | 6;
    } else { // ADF5355
        uint8_t lowfreq = 0;
        if(!this->upperOut) lowfreq = 1;
        this->regs[6] = 1 << 29 | 0xA << 25 | 1 << 24 | (this->divOut) << 21 | ibleed << 13 | (lowfreq) << 10 | (lowfreq) << 6 | 3 << 4 | 6;
    }
    this->regs[7] = 4 << 26 | 1 << 25 | 1 << 7 | 3 << 5 | 7;
    this->regs[8] = 0x102D0428;
    this->regs[9] = 0x15153CC9; // Dirty fixed value !

    uint16_t ADCCLKDIV = ((((double)this->PFD / 100000) -2)/4)+1;

    if (ADCCLKDIV > 255) ADCCLKDIV = 255;

    this->regs[10] = 3 << 22 | ADCCLKDIV << 6 | 1 << 5 | 1 << 4 | 10;
    this->regs[11] = 0x0061300B;
    this->regs[12] = 0x1041C;
}

// Only valid for PFD < 75MHz
void ADFParam::setBlobs(uint8_t type) {
    // Set init blobs
    uint8_t k = 0;
    for(uint8_t i = 0; i < 13; i++) {
        if(i == 12) {
            k = 1;
            blob[i][0] = WAITCYCLE_CMD;
            blob[i][1] = 0x00;
            blob[i][2] = 0x00;
            blob[i][3] = 0x20;
            blob[i][4] = 0x88;
        }
        blob[i+k][0] = (type == 0) ? CONFIGREF_CMD : CONFIGGEN_CMD;
        blob[i+k][1] = (regs[12-i] >> 24) & 0x000000FF;
        blob[i+k][2] = (regs[12-i] >> 16) & 0x000000FF;
        blob[i+k][3] = (regs[12-i] >> 8) & 0x000000FF;
        blob[i+k][4] = (regs[12-i]) & 0x000000FF;
    }
    k = 0;
    // Set update blob
    for(uint8_t i = 0; i < 8; i++) {
        if(i == 7) {
            k = 1;
            upblob[i][0] = WAITCYCLE_CMD;
            upblob[i][1] = 0x00;
            upblob[i][2] = 0x00;
            upblob[i][3] = 0x20;
            upblob[i][4] = 0x88;
        }
        uint32_t r = 0;
        switch(i) {
        case 0 :
            r = regs[10];
            break;
        case 1:
            r = regs[6];
            break;
        case 2:
            r = CNTRSTEN(regs[4]);
            break;
        case 3:
            r = regs[2];
            break;
        case 4:
            r = regs[1];
            break;
        case 5:
            r = AUTOCALDIS(regs[0]);
            break;
        case 6:
            r = CNTRSTDIS(regs[4]);
            break;
        case 7:
            r = AUTOCALEN(regs[0]);
            break;
        }
        upblob[i+k][0] = (type == 0) ? CONFIGREF_CMD : CONFIGGEN_CMD;
        upblob[i+k][1] = (r >> 24) & 0x000000FF;
        upblob[i+k][2] = (r >> 16) & 0x000000FF;
        upblob[i+k][3] = (r >> 8) & 0x000000FF;
        upblob[i+k][4] = (r) & 0x000000FF;
    }

    for(uint8_t i = 0; i < 14; i++) {
        for(uint8_t j = 0; j < 5; j++) {
            fblob[i * 5 + j] = blob[i][j];
        }
    }

    for(uint8_t i = 0; i < 9; i++) {
        for(uint8_t j = 0; j < 5; j++) {
            fupblob[i * 5 + j] = upblob[i][j];
        }
    }

}
