#include "mainwindow.h"
#include "ui_mainwindow.h"

// FTDI library
//#include <windows.h>
//#include <ftdi/ftd2xx.h>

#include <string>
#include <iostream>
#include <fstream>

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow)
{
    ui->setupUi(this);

    /* Init exernal windows and threads */
    this->smithwin = new SmithWin();
    this->myVNA = new VNAThread();

    /* Connect custom signals */
    connect(this->myVNA, SIGNAL(sendData(const DataVector*)), this, SLOT(getData(const DataVector*)));
    connect(this->ui->bStop, SIGNAL(clicked()), this->myVNA, SLOT(stopProcess()));
    connect(this->ui->oglDisp, SIGNAL(sendMarkPoint(const QPointF*)), this, SLOT(setMarker(const QPointF*)));

    /* Init marker table */
    ui->t_markers->setRowCount(MAXMARKERS);
}

MainWindow::~MainWindow()
{
    delete ui;
}

void MainWindow::on_bConnect_clicked()
{
    this->vnaFTDI = new FTDI();
    FTDIErr err = vnaFTDI->init();
    if(err == OK) {
        ui->tStatus->setText("Connected");
        err = vnaFTDI->set245Mode();
        if(err == OK) {
            err = vnaFTDI->setLatency(2);
            if(err == OK) {
                ui->tStatus->setText("FTDI Ready");
                ui->bStart->setEnabled(true);
            } else {
                ui->tStatus->setText("Can't set Latency");
            }
        } else {
            ui->tStatus->setText("Can't set 245 Mode");
        }
    } else {
        ui->tStatus->setText("Not Found");
    }
}

void MainWindow::on_bStart_clicked()
{
    /*
     * Create PLL instance
     * Oscillator : 50MHz
     * VCOMIN : 3400MHz
     * VCOMAX : 6800MHz
     * PFD : 50MHz
     */
    pllRef = new ADF4355(24*1e6, (uint64_t)((double)3.4*(double)1e9), (uint64_t)((double)6.8*(double)1e9), 48*1e6, 0);
    pllRef->initFreqRange((uint64_t)convertToValues(ui->tFreqStart->text()),
                          (uint64_t)convertToValues(ui->tFreqStop->text()),
                          (uint16_t)ui->tNbStep->text().toDouble());

    pllGen = new ADF4355(24*1e6, (uint64_t)((double)3.4*(double)1e9), (uint64_t)((double)6.8*(double)1e9), 48*1e6, 1);
    pllGen->initFreqRange((uint64_t)convertToValues(ui->tFreqStart->text())+1*1e6,
                          (uint64_t)convertToValues(ui->tFreqStop->text())+1*1e6,
                          (uint16_t)ui->tNbStep->text().toDouble());
	
    nbPoints = (uint64_t)ui->tNbStep->text().toDouble();

	this->measIndex = 0;
    if(shortcal == 0) {
        this->ui->oglDisp->setSize(nbPoints);
        this->smithwin->schart->setSize(nbPoints);
        this->shortcal = new std::complex<double>[nbPoints];
        this->opencal = new std::complex<double>[nbPoints];
        this->loadcal = new std::complex<double>[nbPoints];
    }

//    ui->t_markers->clearContents();
//    if(markers != 0) delete []markers;
//    markers = new Marker[MAXMARKERS];

    ui->bShort->setEnabled(false);
    ui->bOpen->setEnabled(false);
    ui->bLoad->setEnabled(false);

    // Set the thread parameters
    myVNA->setParam(1000, this->vnaFTDI, this->pllRef, this->pllGen, (uint64_t)ui->tNbStep->text().toDouble());
    myVNA->start();
}

void MainWindow::getData(const DataVector * c1) {
    // Get the analytic signal
    analytic(c1->sigRef, FFTSIZE, 1);
    analytic(c1->sigDut, FFTSIZE, 1);

    // Use PLL to track the center Freq
    std::complex<double> * lockedRef = 0;
    double reffreq = pll(reinterpret_cast<std::complex<double>*>(c1->sigRef), lockedRef, FFTSIZE);
	
    const double FS = 20e6;
    const double PI = 3.14;
	
    double freq = (reffreq/(2*PI))*FS;

    this->ui->tDebug->setText(QString::number(freq));
    
	// Convolve by the sine from the PLL (1 point DFT)
    std::complex<double> ref = convol1pt(lockedRef, reinterpret_cast<std::complex<double>*>(c1->sigRef), 500);
    std::complex<double> mea = convol1pt(lockedRef, reinterpret_cast<std::complex<double>*>(c1->sigDut), 500);

    // Need to add calibration like in matlab code

    switch(mode) {
        case SHORT:
            shortcal[measIndex] = mea/ref;
            shortcalval = true;
            break;
        case OPEN:
            opencal[measIndex] = mea/ref;
            opencalval = true;
            break;
        case LOAD:
            loadcal[measIndex] = mea/ref;
            loadcalval = true;
            break;
    }

    bool calvalid = shortcalval && opencalval && loadcalval;
    std::complex<double> S11 = mea/ref;
    if(calvalid) { // We have all calibration points !
        std::complex<double> v = (loadcal[measIndex]*(opencal[measIndex]+shortcal[measIndex])-(double)2.0*opencal[measIndex]*shortcal[measIndex]);
        std::complex<double> a = (loadcal[measIndex]*(opencal[measIndex]-shortcal[measIndex]))/v;
        std::complex<double> b = ((double)2.0*loadcal[measIndex]-opencal[measIndex]-shortcal[measIndex])/v;
        std::complex<double> c = (opencal[measIndex]-shortcal[measIndex])/v;
        S11 = (a - c * S11)/(b * S11 - (double)1.0);
    }

    // Crappy test
    MyGLWidget * disp = this->ui->oglDisp;

    float xfactor = (float)disp->width()/(float)disp->size;
    float ysize = (float)disp->height();
    float offset = (float)disp->height()/2;

    double magnitude = mag2db(std::abs(S11));
    double ymag = ysize/(double)200;
    magnitude = magnitude * ymag + offset;

    double phase = std::atan2(std::imag(S11), std::real(S11))*(180.0/PI);
    double yphase = (ysize/(double)2)/(double)180;
    phase = phase * yphase + offset;

    disp->mag[measIndex].setX((float)measIndex*xfactor);
    disp->mag[measIndex].setY(magnitude);
    disp->ang[measIndex].setY(phase);
    disp->ang[measIndex].setX((float)measIndex*xfactor);

    disp->refmag[measIndex].setX((float)measIndex*xfactor);
    double refpow = 10.0+20.0*std::log10(std::pow(std::abs((ref/(double)65536)), 2)/0.1);
    disp->refmag[measIndex].setY(mag2db(refpow));

    disp->repaint();

    // Send data to SmithWin
    if(!smithHide) {
        scwid->SList[measIndex] = S11;
        scwid->repaint();
    }

    // Scan marker list
    for(uint32_t i = 0; i<markCnt;i++) {
        if(markList[i] == measIndex) {
            double freqMark = myVNA->genParam->freqs[measIndex];
            ui->t_markers->setItem(i, 0, new QTableWidgetItem(QString::number(freqMark/1.0e6)));
            ui->t_markers->setItem(i, 1, new QTableWidgetItem(QString::number(i)));
            ui->t_markers->setItem(i, 2, new QTableWidgetItem(QString::number(mag2db(std::abs(S11)))));
            ui->t_markers->setItem(i, 3, new QTableWidgetItem(QString::number(std::atan2(std::imag(S11), std::real(S11))*(180.0/PI))));
            std::complex<double> Zl = -1.0 * 50.0 * (S11+1.0)/(S11-1.0);
            QString Zstr = QString("%1 %2j").arg(QString::number(Zl.real()), QString::number(Zl.imag()));
            ui->t_markers->setItem(i, 4, new QTableWidgetItem(Zstr));
            QString Sstr = QString("%1 %2j").arg(QString::number(S11.real()), QString::number(S11.imag()));
            ui->t_markers->setItem(i, 5, new QTableWidgetItem(Sstr));
            break;
        }
    }

    if(this->measIndex == nbPoints-1) {
        this->measIndex = 0;
        ui->bShort->setEnabled(true);
        ui->bOpen->setEnabled(true);
        ui->bLoad->setEnabled(true);
        mode = MEASURE;
    } else {
        this->measIndex++;
    }
}

void MainWindow::setMarker(const QPointF *p)
{
    MyGLWidget * disp = this->ui->oglDisp;
    float xfactor = (float)disp->width()/(float)disp->size;
    uint32_t x = p->x()/xfactor;
    if(markCnt == MAXMARKERS-1) {
        markCnt = 0;
    }
    markList[markCnt] = x;
    markCnt++;
}

double MainWindow::convertToValues(const QString& input) {
    QRegExp r = QRegExp("^(.+)([nµumKMG])$");
    r.indexIn(input);

    if ( r.captureCount() == 2 ) {

        QString strValue = r.cap(1);
        QString unit = r.cap(2);

        bool ok = false;
        double value = strValue.toDouble(&ok);
        if ( !ok ) {
            return input.toDouble();
        }

        if ( unit == "n" ) {
            return (value*1e-9);
        } else
        if ( unit == "u" || unit == "µ" ) {
            return (value*1e-6);
        } else
        if ( unit == "m" ) {
            return (value*1e-3);
        } else
        if ( unit == "K" ) {
            return (value*1e3);
        } else
        if ( unit == "M" ) {
            return (value*1e6);
        } else
        if ( unit == "G" ) {
            return (value*1e9);
        }

        return value;
    }

    return 0;
}

void MainWindow::on_bStop_clicked()
{
    delete pllGen;
    delete pllRef;
}

void MainWindow::on_bShort_clicked()
{
    mode = SHORT;
}

void MainWindow::on_bOpen_clicked()
{
    mode = OPEN;
}

void MainWindow::on_bLoad_clicked()
{
    mode = LOAD;
}

void MainWindow::on_bSmith_clicked()
{
    if(smithHide) {
        this->smithwin->show();
        smithHide = false;
        scwid = smithwin->schart;
    }
}
