`timescale 1ns / 1ps
`default_nettype none

module main_fsm(clk,
		  DIN, IVALID, IREADY,
		  DOUT, OVALID, OREADY,
		  
		  ADF4355_DOUT, ADF4355_SEND, ADF4355_READY,
		  ADF5355_SEND, ADF5355_READY,
		  LT2271_DOUT, LT2271_SEND, LT2271_READY,
		  
		  ADC_BUSY, ADC_DVALID, ADC_READ, ADC_NSAMP, ADC_START, ADC_DATA,
		  SIG_DRIVER, STATUS);
		  
  input wire clk;
  
  input wire [7:0]DIN;
  input wire IVALID;
  output reg IREADY = 0;
  
  output reg [7:0]DOUT = 0;
  output reg OVALID = 0;
  input wire OREADY;
  
  // ADF4355 signals
  output reg [31:0]ADF4355_DOUT;
  output reg ADF4355_SEND;
  input wire ADF4355_READY;
  
	// ADF5355 signals
  output reg ADF5355_SEND;
  input wire ADF5355_READY;
  
  // LT2271 signals
  output reg [15:0]LT2271_DOUT;
  output reg LT2271_SEND;
  input wire LT2271_READY;
  
  input wire ADC_DVALID;
  input wire ADC_BUSY;
  output reg ADC_READ = 0;
  output reg [12:0]ADC_NSAMP = 13'd4000;
  output reg ADC_START = 0;
  input wire [7:0]ADC_DATA;
  
  // Misc CONTROL
  output reg [7:0]SIG_DRIVER = 0;
  input wire [7:0]STATUS;
  
  // Commands List
  localparam ECHOTEST			= 8'd255;
  localparam ADF4355_SPI		= 8'd1;
  localparam WAIT				= 8'd2;
  localparam LT2271_SPI			= 8'd3;
  localparam GET_SAMPLES		= 8'd4;
  localparam ADF5355_SPI		= 8'd5;
  localparam SET_CONTROLS		= 8'd6;
  localparam GET_STATUS			= 8'd7;
  
  // FSM Control
  
  localparam IDLE				= 0;
  localparam CMDPARSE			= 1;
  localparam ECHOBACK			= 2;
  localparam ADF4355GETDATA		= 3;
  localparam WAITCYCLE			= 4;
  localparam PUSHADF4355        = 5;
  localparam LT2271GETDATA		= 6;
  localparam PUSHLT2271			= 7;
  localparam STARTSAMP			= 8;
  localparam WAITSAMP			= 9;
  localparam GETSAMP			= 10;
  localparam GETCTRLVAL			= 11;
  localparam SENDSTATUS			= 12;
  
  reg [4:0]state = IDLE;
  
  reg [31:0]count = 0; // used to count cycle of command
  
  reg [31:0]gen_reg = 0; // used to store data
  
  reg REFPLL = 0;
  
  always @(posedge clk)
  begin
	ADF4355_SEND <= 0;
	ADF5355_SEND <= 0;
	LT2271_SEND <= 0;
  
	case(state)
		IDLE: begin
			count <= 0;
			gen_reg <= 0;
			if(IVALID) begin
				case(DIN)
					ECHOTEST: state <= ECHOBACK;
					ADF4355_SPI: begin
						state <= ADF4355GETDATA;
						REFPLL <= 1;
					end
					WAIT: state <= WAITCYCLE;
					LT2271_SPI: state <= LT2271GETDATA;
					GET_SAMPLES: state <= STARTSAMP;
					ADF5355_SPI: begin
						state <= ADF4355GETDATA;
						REFPLL <= 0;
					end
					SET_CONTROLS: state <= GETCTRLVAL;
					GET_STATUS: state <= SENDSTATUS;
				endcase
			end
		end
		
//		ECHOBACK: begin
//			if(OREADY && IVALID) begin
//                state <= IDLE;
//				DOUT <= DIN;
//				OVALID <= 1;
//			end
//		end
		
		ADF4355GETDATA: begin
			if(IVALID && ((ADF4355_READY && REFPLL) || (ADF5355_READY && !REFPLL))) begin
				count <= count + 1;
				gen_reg <= {gen_reg[23:0], DIN};
				if(count == 3) begin
					state <= PUSHADF4355;
				end
			end
		end

        PUSHADF4355: begin
            ADF4355_DOUT <= gen_reg;
				if(REFPLL) ADF4355_SEND <= 1;
				else ADF5355_SEND <= 1;
            state <= IDLE;
        end
		
		WAITCYCLE: begin
			if(IVALID) begin
				count <= count + 1;
				if(count <= 3) begin
					gen_reg <= {gen_reg[23:0], DIN};
				end else if(count >= gen_reg) begin
					state <= IDLE;
				end
			end
		end
		
		LT2271GETDATA: begin
			if(IVALID && LT2271_READY) begin
				count <= count + 1;
				gen_reg <= {gen_reg[23:0], DIN};
				if(count == 1) begin
					state <= PUSHLT2271;
				end
			end
		end
		
		PUSHLT2271: begin
			LT2271_DOUT <= gen_reg[15:0];
			LT2271_SEND <= 1;
			state <= IDLE;
		end
		
		STARTSAMP: begin
			if(ADC_BUSY) state <= WAITSAMP;
		end
		
		WAITSAMP: begin
			if(!ADC_BUSY && ADC_DVALID) begin
				state <= GETSAMP;
			end
		end
		
		GETSAMP: begin
			if(!ADC_DVALID) state <= IDLE;
		end
		
		GETCTRLVAL: begin
			if(IVALID) begin
				SIG_DRIVER <= DIN;
				state <= IDLE;
			end
		end
		
		SENDSTATUS: begin
			if(OREADY) begin
				state <= IDLE;
			end
		end
	endcase
  end
  
  always @*
  begin
		IREADY = 1;
		DOUT = 0;
		OVALID = 0;
		ADC_READ = 0;
		ADC_START = 0;
		
		case(state)
			PUSHADF4355: IREADY = 0;
			
			ECHOBACK: begin
				if(!IVALID || !OVALID) begin
					IREADY = 0;
				end
			end
			
			ADF4355GETDATA: begin
				if(!IVALID || (!ADF4355_READY && REFPLL) || (!ADF5355_READY && !REFPLL)) begin
					IREADY = 0;
				end
			end
			
			PUSHLT2271: IREADY = 0;
			
			LT2271GETDATA: begin
				if(!IVALID || !LT2271_READY) begin
					IREADY = 0;
				end
			end
			
			WAITCYCLE: begin
				if(count >= 3) begin
					IREADY = 0;
				end
			end
			
			STARTSAMP: begin
				ADC_START = 1;
				IREADY = 0;
			end
			
			WAITSAMP: begin
				IREADY = 0;
			end
			
			GETSAMP: begin
				IREADY = 0;
				if(ADC_DVALID && OREADY) begin
					DOUT = ADC_DATA;
					OVALID = ADC_DVALID;
					ADC_READ = 1;
				end
			end
			
			SENDSTATUS: begin
				IREADY = 0;
				if(OREADY) begin
					DOUT = STATUS;
					OVALID = 1;
				end
			end
		endcase
  end
  
  
//	wire [35:0]CONTROL0;
//
//	chipscope_icon chipcon(.CONTROL0(CONTROL0));
//
//	chipscope_ila rxftdi (
//		.CONTROL(CONTROL0),
//		.CLK(clk),
//		.DATA({ADF4355_READY, ADF5355_READY, IREADY, REFPLL, DIN}),
//		.TRIG0(state != IDLE)
//	);
  
endmodule
